#ifndef GR_ACL_H
#define GR_ACL_H
#endif
#include <linux/grsecurity.h>
#include <linux/grdefs.h>

/* * * * * * * * * * * * * * * * * * * * *
 * GrSecurity ACL System
 * Main header file
 * Purpose: define most gracl data structures 
 * * * * * * * * * * * * * * * * * * * * */

/* Major status information */

#define GR_VERSION  "grsecurity 1.9.5"

#ifndef CONFIG_GR_MAXTRIES
#define CONFIG_GR_MAXTRIES 3
#endif

#ifndef CONFIG_GR_TIMEOUT
#define CONFIG_GR_TIMEOUT 30
#endif

enum {

	SHUTDOWN = 0,
	ENABLE = 1,
	GOD = 2,
	LEARN = 3,
};

/* Internal status definitions
 * gr_set_abel() in kernel/gracl.c
 * */
enum {
	GR_SET_INHERIT,
	GR_SET_PROC
};

/* Password setup definitions
 * kernel/grhash.c */
enum {
	GR_PW_LEN = 128,
	GR_SALT_LEN = 16,
	GR_SHA_LEN = 20,
};

/* Begin Data Structures */

struct admin_pw {
	unsigned char salt[GR_SALT_LEN];
	unsigned char sum[GR_SHA_LEN];	/* 160-bit SHA hash of the password */
};

struct acl_subject_label {
	ino_t inode;
	kdev_t device;
	__u16 mode;
	kernel_cap_t cap_raise;
	kernel_cap_t cap_lower;
	kernel_cap_t cap_raise_inherit;
	kernel_cap_t cap_lower_inherit;
	/* Keeps track of the capabilities to raise and lower. 
	 * Have to store it this way because users my change the default
	 * capabilities, and we want to change with them(but still
	 * keep our custom raised/lowered caps) */
	unsigned long pos;
	/* .. used for process object
	   hashing. Position (i.e.
	   subject acl number 1..n)
	   The reason positition is part of the hash is that multiple
	   processes may have acls for the same file, and we need some way
	   to preventa  collision from occurring in that case,
	   BUT subject inodes/devices may change because of create/delete
	   operations, so position is the only thing that won't change but
	   will still identify the acl subject
	 */
};

struct acl_object_label {
	ino_t inode;
	kdev_t device;
	__u16 mode;
	unsigned long parent_pos;
};

struct name_entry {
	ino_t inode;
	kdev_t device;
	__u8 mode;
	char *name;
};

struct acl_subject_db {
	struct acl_subject_label *s_hash;
	unsigned long s_size;
	unsigned long s_used;
};

struct acl_object_db {
	struct acl_object_label *o_hash;
	unsigned long o_size;
	unsigned long o_used;
};

struct name_db {
	struct name_entry *n_hash;
	unsigned long n_size;
	unsigned long n_used;
};

struct temp_file_label {
	ino_t inode;
	kdev_t device;
	unsigned long mode;
};

struct acl_info {
	ino_t inode;
	kdev_t device;
	unsigned long pos;
};

struct temp_socket_label {
	unsigned int port;
	__u8 mode;
};

struct label_info {
	ino_t inode;
	kdev_t device;
};

/* Userspace Grsecurity ACL data structures */
struct user_acl_subject_label {
	char * filename;
	unsigned short filename_len;
	ino_t inode;
	kdev_t device;
	__u16 mode;
	__u32 cap_raise;
	__u32 cap_lower;
	__u32 cap_raise_inherit;
	__u32 cap_lower_inherit;
	unsigned long pos;
	
	/* next 3 structures not used */

        struct user_acl_object_label *proc_object;
        struct user_acl_subject_label *prev;
        struct user_acl_subject_label *next;
};

struct user_acl_subject_db {
	struct user_acl_subject_label **s_table;
	unsigned long s_entries;	/* Number of entries in table */
};

struct user_acl_object_label {
	char * filename;
	unsigned short filename_len;
	ino_t inode;
	kdev_t device;
	__u16 mode;
	unsigned long parent_pos;
	
	/* next two structures not used */

 	struct user_acl_object_label *prev;
        struct user_acl_object_label *next;
};

struct user_acl_object_db {
	struct user_acl_object_label **o_table;
	unsigned long o_entries;
};

struct gr_arg {
	struct user_acl_object_db obj_db;
	struct user_acl_subject_db subj_db;
	unsigned char pw[GR_PW_LEN];
	__u16 mode;

};

/* End Data Structures Section */

static __inline__ unsigned long
fhash(const ino_t ino, const kdev_t dev, const unsigned long sz)
{
	return (ino ^ (dev * 797)) % sz;
}

static __inline__ unsigned long
phash(const ino_t ino, const kdev_t dev,
      const unsigned long parent_pos, const unsigned long sz)
{
	/* Okay...parent_pos is the subject acl # (0..n-1 for n subject acls)
	 * We reuse fhash here for the parent_pos portion of phash() for
	 * simplicity's sake */
	return (fhash(ino, dev, ~0UL) ^ fhash((ino_t) (parent_pos), 0, ~0UL))
	    % sz;
}

static __inline__ unsigned long
nhash(const char *name, const unsigned long sz)
{
	return full_name_hash(name, strlen(name)) % sz;
}
